const pos = require('pos')

const C = require('./constraints')

// Hackery to deal with limitations of the parser and typos in the puzzle pdf...
const sanitize = (sentence) => {
  if(sentence === 'The first person on the right likes Swimming.')
    return 'The first on the right likes Swimming.'
  return sentence
}

const tag = (sentence) => {
  let words = new pos.Lexer().lex(sanitize(sentence))
  // TODO: Extend lexicon with the categories?
  let tagger = new pos.Tagger()
  return tagger.tag(words)
}

const tagAllClues = (puzzle) => {
  puzzle.clues = puzzle.clues.map(c => { return {raw: c, tagged: tag(c)} } )
  return puzzle
}

const parse_ = (word, words, tree) => {
  if(!word) {
    if(words.length === 0) return tree
    word = words.shift()
  }

  switch(word[1]) {
  case 'VB':
  case 'VBZ': {
    let left = tree
    let right = parse_(words.shift(), words, [])
    if(left.length === 0 && right.length === 0) {
      return [[word[0], 'NN']] // Likely a POS tag error. e.g. ['Boost', 'VB'] when 'Boost' refers to the name of a candy bar.
    } else {
      return [word, left, right]
    }
  }
  case 'WP': {
    let cursor = 0
    let numVerbsFound = 0
    for(;;) { // loop until you find a second verb
      if(words.length === 0) break
      switch(words[cursor][1]) {
      case 'VB':
      case 'VBZ': {
        ++numVerbsFound
        if(numVerbsFound > 1) {
          let left = word
          let right = parse_(words.shift(), words.slice(0, cursor-1), [])
          tree.push([word, right])
          return parse_(null, words.slice(cursor-1), tree)
        }
        break
      }
      case '.': {
        tree.push([word, parse_(words.shift(), words, [])])
        return tree
      }
      }
      ++cursor
      if(cursor >= words.length) break
    }
    return tree
  }
  case 'IN': {
    let cursor = 0
    for(;;) { // loop until you find another preposition or a verb
      switch(words[cursor][1]) {
      case 'VB':
      case 'VBZ':
      case 'IN': {
        // End the phrase and continue processing the next one
        tree.push([word, parse_(words.shift(), words.slice(0, cursor-1), [])])
        return parse_(null, words.slice(cursor-1), tree)
      }
      }
      ++cursor
      if(cursor >= words.length) break
    }
    let prepPhrase = [word, parse_(words.shift(), words, [])]
    tree.push(prepPhrase)
    return tree
  }
  case 'DT': {
    return parse_(words.shift(), words, tree)
  }
  case '.': {
    return tree
  }
  default:
    tree.push(word)
    return parse_(words.shift(), words, tree)
  }
}

const parse = (words) => {
  return parse_(null, words, [])
}

const parseAllClues = (puzzle) => {
  puzzle.clues = puzzle.clues.map(c => { return {raw: c.raw, tagged: c.tagged, parsed: parse(c.tagged.slice()) } } )
  return puzzle
}


// The overall goal here is to reduce / simplify the tree recursively until we get to one of the clues.
// What filtering/reordering/relabeling rules can I create?
const analyze = (tree) => {
  if(!tree || tree.length === 0) { return }
  let token = tree.shift()
  if(Array.isArray(token[1])) { // Compound statement, skip this level of the tree
    let child = analyze(token[1])
    return [child].concat(analyze(tree))
  }

  switch(token[1]) {
  case 'JJ': {
    return [C.NextTo, analyze(tree)]
  }
  case 'IN': {
    return analyze(tree)
  }
  case 'VBG':
  case 'VBN':
  case 'NN':
  case 'NNS':
  case 'NNP': {
    let child = token[0]
    // is there another sibling? and does it start with a 'WP' or an 'IN'?
    if(tree[0] && tree[0][0] && tree[0][0][1] === 'WP') {
      let phrase = tree[0]
      return analyze(phrase[1])
    }

    if(tree[0] && tree[0][0] && tree[0][0][1] === 'IN') {
      let phrase = tree[0]
      if(tree.length === 2) {
        let a = analyze(tree[0][1])
        let b = analyze(tree[1][1])
        switch(a) {
        case 'left': {
          return [C.Left, undefined, b]
        }
        case 'right': {
          return [C.Right, undefined, b]
        }
        default: {
          return null
        }
        }
      }

      switch(child) { // Special words
      case 'first': {
        switch(analyze(tree[0][1])) {
          case C.Left: return 1
          case C.Right: return 5
          default: return undefined
        }
      }
      default: {
        return analyze(phrase[1])
      }
      }
    }

    if(tree.length > 0 && tree[0] && tree[0][1] === 'JJ' && tree[0][0] === 'next') {
      return analyze(tree)
    }

    // Looking for compound nouns
    if(tree.length > 0) {
      for(let i = 0; i < tree.length; ++i) {
        let posTag = tree[i][1]
        if(posTag === 'NN' || posTag === 'NNS' || posTag === 'NNP' || posTag === 'VBG') {
          child = `${child} ${tree[i][0]}`
        } else {
          break
        }
      }
    }

    return child
  }
  case 'VB': {
    let left = analyze(tree.shift())
    let right = analyze(tree.shift())
    return [left, right]
  }
  case 'VBZ': {
    let left = analyze(tree.shift())
    let right = analyze(tree.shift())
    return [C.Is, left, right]
  }
  case 'WP': {
    return analyze(tree.shift())
  }
  default: {
    return analyze(tree)
  }
  }
}

const analyzeAllClues = (puzzle) => {
  puzzle.clues = puzzle.clues.map(c => { return {raw: c.raw, tagged: c.tagged, parsed: c.parsed, analyzed: analyze(JSON.parse(JSON.stringify(c.parsed))) } } )
  return puzzle
}

const simplify = (constraint) => {
  if(Array.isArray(constraint[0])) {
    let simplified = constraint.map(c => {
      return Array.isArray(c) ? simplify(c) : c
    })
    if(simplified[0] === null) {
      return simplified[1]
    }
    return simplified
  }

  switch(constraint[0]) {
  case C.Is: {
    if(Array.isArray(constraint[1])) {
      constraint[1] = simplify(constraint[1])
      switch(constraint[1][0]) {
      case C.Left: {
        constraint[1] = constraint[1].slice(1).map(word => {
          switch(word) {
            case 'first':
              return 1
            case 'second':
              return 2
            case 'third':
              return 3
            case 'fourth':
              return 4
            case 'fifth':
              return 5
            case 'middle':
              return 2
            case 'last':
              return 5
            default:
              return 0
          }
        }).reduce((acc, x) => acc+x, 0)
        break
      }
      case C.Right: {
        let position = constraint[1].slice(1).map(word => {
          switch(word) {
            case 'first':
              return 5
            case 'second':
              return 4
            case 'third':
              return 3
            case 'fourth':
              return 2
            case 'fifth':
              return 1
            case 'middle':
              return 4
            case 'last':
              return 1
            default:
              return 0
          }
        }).reduce((acc, x) => acc+x, 0)
        constraint[1] = position === 0 ? C.Right : position
        break
      }
      }
    }
    if(Array.isArray(constraint[2])) {
      constraint[2] = simplify(constraint[2])
      switch(constraint[2][0]) {
      case C.Left: {
        constraint[0] = C.Left
        constraint[2] = constraint[2].filter(c => c !== C.Left)[0]
      }
      }
    }

    if(!constraint[1]) {
      return constraint[2]
    }

    if(!constraint[2]) {
      return constraint[1]
    }

    if(constraint[1] === C.Right) {
      return [C.Right, constraint[2][2], constraint[2][1]]
    }

    if(constraint[1][0] === C.NextTo) {
      return [C.NextTo, constraint[1][1], constraint[2]]
    }

    if(constraint[2][0] === C.NextTo) {
      return [C.NextTo, constraint[1], constraint[2][1]]
    }

    return constraint
  }
  case C.Left: {
    return constraint.filter(c => c)
  }
  case C.NextTo: {
    let simplified = simplify(constraint[1])
    return simplified ? [C.NextTo, simplified] : C.NextTo
  }
  default: {
    let reduced = constraint.filter(c => c)
    return reduced.length === 1 ? reduced[0] : reduced
  }
  }
}

const simplifyAllClues = (puzzle) => {
  puzzle.clues = puzzle.clues.map(c => { return {raw: c.raw, tagged: c.tagged, parsed: c.parsed, analyzed: c.analyzed, simplified: simplify(JSON.parse(JSON.stringify(c.analyzed.slice()))) } } )
  return puzzle
}


module.exports = {
  tag: tagAllClues,
  parse: parseAllClues,
  analyze: analyzeAllClues,
  simplify: simplifyAllClues,
}

