const C = require('./constraints')

const normalizeWord = (word) => {
  // Note: Some hackery to deal with typos in the puzzle pdf. Eventually this should be expanded to use edit distance based on the category (or really, moved into semantic analysis...)
  if(word === 'Ten_pin bowlings') word = 'ten_pin_bowling'
  if(['Swimming', 'Hamster', 'Swimmming', 'Puppy', 'Skiing', 'Rabbit', 'Purple', 'Horse riding', 'Horse_riding', 'Blue', 'Lilac', 'Pink', 'Green', 'Horse', 'Tennis'].includes(word)) word = word.toLowerCase()
  if(word === 'middle') return 3
  if(word === 'first person') return 1
  return word.replace(/[ -]/g, '_')
}

const serializeConstraint = (constraint) => {
  if(constraint[1].replace) constraint[1] = normalizeWord(constraint[1])
  if(constraint[2].replace) constraint[2] = normalizeWord(constraint[2])
  switch(constraint[0]) {
  case C.Between: {
    return `${constraint[2]} < ${constraint[1]} /\\ ${constraint[1]} < ${constraint[3]}`
  }
  case C.Is: {
    return `${constraint[1]} = ${constraint[2]}`
  }
  case C.Left: {
    return `${constraint[1]} < ${constraint[2]}`
  }
  case C.Right: {
    return `${constraint[1]} > ${constraint[2]}`
  }
  case C.LeftImmediately: {
    return `${constraint[1]} = ${constraint[2]} - 1`
  }
  case C.RightImmediately: {
    return `${constraint[1]} = ${constraint[2]} + 1`
  }
  case C.NextTo: {
    return `abs(${constraint[1]} - ${constraint[2]}) = 1`
  }
  default:
    throw `Unknown constraint type: ${constraint[0]}`
  }
}

const renderModel = (p) => {
  let output = 'include "globals.mzn";\n'

  Object.keys(p.categories).forEach(key => {
    output += `array[1..5] of var 1..5: ${key};\n`
    p.categories[key].forEach(variant => {
      output += `var int: ${normalizeWord(variant)};\n`
    })
  })

  output += '\nconstraint\n'
  output += Object.keys(p.categories).map(key => `  inverse(${key}, [${p.categories[key].map(normalizeWord)}]) /\\\n`).join('')
  output += p.clues.map(clue => `  ${serializeConstraint(clue.simplified)}`).join(' /\\\n')
  output += ';\n\n'

  output += 'solve satisfy;\n'

  output += '\noutput\n[\n'
  Object.keys(p.categories).forEach(key => {
    output += `"${key} ", show([${p.categories[key].map(normalizeWord)}]), "\\n",\n`
  })
  output += '];\n'

  return output
}

const launchSolver = (model) => {
  const { execSync } = require('child_process')
  try {
    const stdout = execSync('minizinc --solver Gecode --input-from-stdin', { input: model })
    return stdout.toString()
  } catch(error) {
    console.error(`error: ${error.message}`)
    return
  }
}

const renderSolution = (parsedPuzzle, solverOutput) => {
  if (solverOutput.includes("UNSATISFIABLE")) {
    return solverOutput
  }

  const orderings = solverOutput.split('\n')
                    .filter(line => line.indexOf('-') < 0 && line.length > 0)
                    .map(line => line.replace(/[,\[\]]/g, ''))
                    .map(line => line.split(' '))
                    .map(line => { let obj = {}; obj[line[0]] = line.slice(1).map(i => parseInt(i)); return obj})
                    .reduce((acc, obj) => {return {...acc, ...obj}})

  return Object.keys(parsedPuzzle.categories).map(key => {
    let sorted = [null, null, null, null, null]
    for(let i = 0; i < 5; ++i) {
      sorted[orderings[key][i]-1] = parsedPuzzle.categories[key][i]
    }

    let res = {}
    // strip the disambiguating prefix that was added to turn numbers into valid identifiers for minizinc
    res[key] = sorted.map(el => el.indexOf(key) === 0 ? el.substring(key.length) : el)
    return res
  })
  .reduce((acc, obj) => {return {...acc, ...obj}})
}

const extractPuzzle = require('./pdf')
const {tag, parse, analyze, simplify} = require('./parse')
const solve = async (filePath) => {
  const puzzle = simplify(analyze(parse(tag(await extractPuzzle(filePath)))))
  return renderSolution(puzzle, launchSolver(renderModel(puzzle)))
}

module.exports = solve
